using System.Diagnostics;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using System;

namespace Optim_BFGS1
{
	public delegate double MyFxDelegate(int nNumVars, ref double[] fX, ref double[] fParam);
	public delegate string SayFxDelegate();
	
	public class COptimBFGS1
	{
		
		MyFxDelegate m_MyFx;
		
		protected double MyFxEx(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fDeltaX, double fLambda)
		{
			int i;
			double[] fXX = new double[nNumVars];
			
			for (i = 0; i < nNumVars; i++)
			{
				fXX[i] = fX[i] + fLambda * fDeltaX[i];
			}
			
			return m_MyFx(nNumVars, ref fXX, ref fParam);
		}
		
		protected double FirstDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI)
		{
			double fXt, h, Fp, Fm;
			
			fXt = fX[nIdxI];
			h = 0.01 *(1 + Math.Abs(fXt));
			fX[nIdxI] = fXt + h;
			Fp = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt - h;
			Fm = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt;
			return (Fp - Fm) / 2 / h;
		}
		
		protected void GetFirstDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fFirstDerivX)
		{
			
			int i;
			
			for (i = 0; i < nNumVars; i++)
			{
				fFirstDerivX[i] = FirstDeriv(nNumVars, ref fX, ref fParam, i);
			}
		}
		
		protected bool LinSearch_DirectSearch(int nNumVars, ref double[] fX, ref double[] fParam, ref double fLambda, ref double[] fDeltaX, double fInitStep, double fMinStep)
		{
			double F1, F2;
			
			F1 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda);
			
			do
			{
                F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda + fInitStep);
				if (F2 < F1)
				{
					F1 = F2;
					fLambda += fInitStep;
				}
				else
				{
                    F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda - fInitStep);
					if (F2 < F1)
					{
						F1 = F2;
						fLambda -= fInitStep;
					}
					else
					{
						// reduce search step size
						fInitStep /= 10;
					}
				}
			} while (!(fInitStep < fMinStep));
			
			return true;
			
		}
		
		public double CalcOptim(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fToler, double fEpsFx, double fEpsGrad, int nMaxIter, ref int nIter, ref string sErrorMsg, MyFxDelegate MyFx)
		{
			
			int i;
			double fNorm, fLambda;
			bool bStop;
			double T1, T2;
			// Declare local arrays
			int[] Index = new int[nNumVars];
			double[] fGrad1 = new double[nNumVars];
			double[] fGrad2 = new double[nNumVars];
			double[] g = new double[nNumVars];
			double[] d = new double[nNumVars];
			double[] S = new double[nNumVars];
			double[,] Bmat = new double[nNumVars, nNumVars];
			double[,] Mmat = new double[nNumVars, nNumVars];
			double[,] Nmat = new double[nNumVars, nNumVars];
			double[,] MM1 = new double[nNumVars, nNumVars];
			double[,] MM2 = new double[nNumVars, nNumVars];
			double[,] MM3 = new double[nNumVars, nNumVars];
			double[,] MM4 = new double[nNumVars, nNumVars];
			double[,] MM5 = new double[nNumVars, nNumVars];
			double[,] MM6 = new double[nNumVars, nNumVars];
			double[,] MM7 = new double[nNumVars, nNumVars];
			double[,] MM8 = new double[nNumVars, nNumVars];
			double[,] MM9 = new double[nNumVars, nNumVars];
			double[,] MM10 = new double[nNumVars, nNumVars];
			double[,] MM11 = new double[nNumVars, nNumVars];
			
			
			// set error handler
			try
			{
				
				m_MyFx = MyFx;
				// set matrix B as an indentity
				MatrixLibVb.MatIdentity(ref Bmat);
				
				nIter = 1;
				// calculate initial gradient
				GetFirstDerives(nNumVars, ref fX, ref fParam, ref fGrad1);
				
				// start main loop
				do
				{
					
					nIter++;
					if (nIter > nMaxIter)
					{
						sErrorMsg = "Iter limits";
						break;
					}
					
					MatrixLibVb.MatMultVect(ref Bmat, ref fGrad1, ref S);
					MatrixLibVb.VectMultValInPlace(ref S, - 1);
					
					// test if gradient is shallow enough
					MatrixLibVb.NormalizeVect(ref S);
					
					fLambda = 0.1;
					LinSearch_DirectSearch(nNumVars, ref fX, ref fParam, ref fLambda, ref S, 0.1, 0.00001);
					// calculate optimum fX()
					for (i = 0; i < nNumVars; i++)
					{
						d[i] = fLambda * S[i];
						fX[i] += d[i];
					}
					
					// get new gradient
					GetFirstDerives(nNumVars, ref fX, ref fParam, ref fGrad2);
					for (i = 0; i < nNumVars; i++)
					{
						g[i] = fGrad2[i] - fGrad1[i];
						fGrad1[i] = fGrad2[i];
					}
					
					// test for convergence
					bStop = true;
					for (i = 0; i < nNumVars; i++)
					{
						if (Math.Abs(d[i]) > fToler[i])
						{
							bStop = false;
							break;
						}
					}
					
					if (bStop)
					{
						sErrorMsg = "Lambda * S()";
						break;
					}
					
					// exit if gradient is low
					if (MatrixLibVb.VectNorm(ref g, 2) < fEpsGrad)
					{
						sErrorMsg = "G Norm";
						break;
					}
					
					//-------------------------------------------------
					// Start elaborate process to upgare matrix B
					//
					MatrixLibVb.VectToColumnMat(ref g, ref MM1); // MM1 = g as column matrix
					MatrixLibVb.VectToColumnMat(ref d, ref MM2); // MM2 = d as column matrix
					MatrixLibVb.VectToRowMat(ref g, ref MM3); // MM3 = g^T
					MatrixLibVb.VectToRowMat(ref d, ref MM4); // MM4 = d^T
					MatrixLibVb.MatMultMat(ref MM2, ref MM4, ref MM5); // MM5 = d d^T
					MatrixLibVb.MatMultMat(ref MM4, ref MM1, ref MM6); // MM6 = d^T g
					MatrixLibVb.MatMultMat(ref MM2, ref MM3, ref MM7); // MM7 = d g^T
					MatrixLibVb.MatMultMat(ref MM1, ref MM4, ref MM8); // MM8 = g d^T
					MatrixLibVb.MatMultMat(ref MM3, ref Bmat, ref MM9);
					MatrixLibVb.MatMultMatInPlace(ref MM9, ref MM1); // MM9 = g^T [B] g
					MatrixLibVb.MatMultMat(ref MM7, ref Bmat, ref MM10); // MM10 = d g^T [B]
					MatrixLibVb.MatMultMat(ref Bmat, ref MM8, ref MM11); // MM11 = [B] g d^T
					
					T1 = MM6[0, 0]; // d^T g
					T2 = (1 + MM9[0, 0] / T1) / T1;
					MatrixLibVb.MatMultValInPlace(ref MM5, T2); //
					MatrixLibVb.MatMultValInPlace(ref MM10, 1 / T1);
					MatrixLibVb.MatMultValInPlace(ref MM11, 1 / T2);
					MatrixLibVb.MatAddMatInPlace(ref Bmat, ref MM5);
					MatrixLibVb.MatSubMatInPlace(ref Bmat, ref MM10);
					MatrixLibVb.MatSubMatInPlace(ref Bmat, ref MM11);
					
				} while (true);
			}
			catch(Exception ex)
			{
				sErrorMsg = "Error: " + ex.Message;
			}
           return MyFx(nNumVars, ref fX, ref fParam);
		}
		
	}
	
	
}
